@@ -562,6 +562,7 @@ class Session(object):
         # since our connection has failed retries
         # deregister our keepalive facility
         self.lastpayload = None
+        self.onlogpayload = None
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         with util.protect(WAITING_SESSIONS):
@@ -1050,6 +1051,7 @@ class Session(object):
                                             data=[self.privlevel])
             if response['code']:
                 self.logged = 0
+                self.onlogpayload = None
                 self.logging = False
                 mysuffix = " while requesting privelege level %d for %s" % (
                     self.privlevel, self.userid)
@@ -1126,7 +1128,7 @@ class Session(object):
                                         data=[0x8e, self.privlevel])
 
     def login(self):
-        self.logontries = 5
+        self.logontries = 2
         self._initsession()
         self._get_channel_auth_cap()
 
@@ -1348,7 +1350,7 @@ class Session(object):
                 iserver.pktqueue.append(qent)
                 iserver.process_pktqueue()
                 return
-            if (hasattr(self, 'remseqnumber')
+            if (hasattr(self, 'remseqnumber') and self.remseqnumber is not None
                     and remseqnumber < self.remseqnumber):
                 return -5  # remote sequence number is too low, reject it
             self.remseqnumber = remseqnumber
@@ -1426,7 +1428,7 @@ class Session(object):
             if sid != self.localsid:  # session id mismatch, drop it
                 return
             remseqnumber = struct.unpack("<I", bytes(data[10:14]))[0]
-            if (hasattr(self, 'remseqnumber')
+            if (hasattr(self, 'remseqnumber') and self.remseqnumber is not None
                     and (remseqnumber < self.remseqnumber)
                     and (self.remseqnumber != 0xffffffff)):
                 return
@@ -1842,6 +1844,7 @@ class Session(object):
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         self.logged = 0
+        self.onlogpayload = None
         self.logging = False
         if self._customkeepalives:
             for ka in list(self._customkeepalives):
