@@ -25,9 +25,10 @@ import socket
 import ssl
 import threading
 
-import pyghmi.exceptions as pygexc
 import six
 
+import pyghmi.exceptions as pygexc
+
 try:
     import Cookie
     import httplib
@@ -47,7 +48,7 @@ uploadforms = {}
 class FileUploader(threading.Thread):
 
     def __init__(self, webclient, url, filename, data=None, formname=None,
-                 otherfields=()):
+                 otherfields=(), formwrap=True, excepterror=True):
         self.wc = webclient
         self.url = url
         self.filename = filename
@@ -56,15 +57,22 @@ class FileUploader(threading.Thread):
         self.formname = formname
         self.rsp = ''
         self.rspstatus = 500
+        self.formwrap = formwrap
+        self.excepterror = excepterror
         super(FileUploader, self).__init__()
 
     def run(self):
         try:
             self.rsp = self.wc.upload(
                 self.url, self.filename, self.data, self.formname,
-                otherfields=self.otherfields)
-        except Exception:
+                otherfields=self.otherfields, formwrap=self.formwrap,
+                excepterror=self.excepterror)
             self.rspstatus = self.wc.rspstatus
+        except Exception:
+            try:
+                self.rspstatus = self.wc.rspstatus
+            except Exception:
+                pass
             raise
 
 
@@ -74,10 +82,14 @@ class FileDownloader(threading.Thread):
         self.wc = webclient
         self.url = url
         self.savefile = savefile
+        self.exc = None
         super(FileDownloader, self).__init__()
 
     def run(self):
-        self.wc.download(self.url, self.savefile)
+        try:
+            self.wc.download(self.url, self.savefile)
+        except Exception as e:
+            self.exc = e
 
 
 def get_upload_form(filename, data, formname, otherfields):
@@ -90,18 +102,16 @@ def get_upload_form(filename, data, formname, otherfields):
             data = data.read()
         except AttributeError:
             pass
-        form = (b'--' +
-                BND +
-                '\r\nContent-Disposition: form-data; '
-                'name="{0}"; filename="{1}"\r\n'.format(
-                    formname, filename).encode('utf-8'))
+        form = (b'--' + BND
+                + '\r\nContent-Disposition: form-data; '
+                  'name="{0}"; filename="{1}"\r\n'.format(
+                      formname, filename).encode('utf-8'))
         form += b'Content-Type: application/octet-stream\r\n\r\n' + data
         for ofield in otherfields:
-            form += (b'\r\n--' +
-                     BND +
-                     '\r\nContent-Disposition: form-data; '
-                     'name="{0}"\r\n\r\n{1}'.format(
-                         ofield, otherfields[ofield]).encode('utf-8'))
+            form += (b'\r\n--' + BND
+                     + '\r\nContent-Disposition: form-data; '
+                       'name="{0}"\r\n\r\n{1}'.format(
+                           ofield, otherfields[ofield]).encode('utf-8'))
         form += b'\r\n--' + BND + b'--\r\n'
         uploadforms[filename] = form
         return form
@@ -126,7 +136,6 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
                                             **kwargs)
         except TypeError:
             httplib.HTTPConnection.__init__(self, host, port, **kwargs)
-        self.cert_reqs = ssl.CERT_NONE  # verification will be done ssh style..
         if clone:
             self._certverify = clone._certverify
             self.cookies = clone.cookies
@@ -135,7 +144,11 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
             self._certverify = verifycallback
             self.cookies = {}
             self.stdheaders = {}
-        if '[' not in host and '%' in host:
+        if self._certverify:
+            self.cert_reqs = ssl.CERT_NONE  # use custom validation
+        else:
+            self.cert_reqs = ssl.CERT_REQUIRED  # use standard validation
+        if '[' not in host and '%' in host and 'Host'not in self.stdheaders:
             self.stdheaders['Host'] = '[' + host[:host.find('%')] + ']'
 
     def dupe(self):
@@ -169,12 +182,19 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
         except socket.error:
             pass
         plainsock.connect(addrinfo[4])
-        self.sock = ssl.wrap_socket(plainsock, cert_reqs=self.cert_reqs)
-        # txtcert = self.sock.getpeercert()  # currently not possible
-        bincert = self.sock.getpeercert(binary_form=True)
-        if not self._certverify(bincert):
-            raise pygexc.UnrecognizedCertificate('Unknown certificate',
-                                                 bincert)
+        if self._certverify:
+            self.sock = ssl.wrap_socket(plainsock, cert_reqs=self.cert_reqs)
+            bincert = self.sock.getpeercert(binary_form=True)
+            if not self._certverify(bincert):
+                raise pygexc.UnrecognizedCertificate('Unknown certificate',
+                                                     bincert)
+        else:
+            ctx = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
+            ctx.load_default_certs()
+            ctx.verify_mode = ssl.CERT_REQUIRED
+            ctx.check_hostname = True
+            self.sock = ctx.wrap_socket(plainsock,
+                                        server_hostname=self.thehost)
 
     def getresponse(self):
         try:
@@ -226,10 +246,18 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
             raise
         body = rsp.read()
         if rsp.getheader('Content-Encoding', None) == 'gzip':
-            body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            try:
+                body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            except IOError:
+                # some implementations will send non-gzipped and claim it as
+                # gzip
+                pass
         if rsp.status >= 200 and rsp.status < 300:
-            if body and not isinstance(body, str):
-                body = body.decode('utf8')
+            if body and not isinstance(body, type(u'')):
+                try:
+                    body = body.decode('utf8')
+                except Exception:
+                    body = body.decode('iso-8859-1')
             return json.loads(body) if body else {}, rsp.status
         return body, rsp.status
 
@@ -247,8 +275,10 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
         rsp = webclient.getresponse()
         self._currdl = rsp
         self._dlfile = file
-        for chunk in iter(lambda: rsp.read(16384), ''):
+        chunk = rsp.read(16384)
+        while chunk:
             file.write(chunk)
+            chunk = rsp.read(16384)
         self._currdl = None
         file.close()
 
@@ -259,7 +289,7 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
             self._currdl.getheader('content-length'))
 
     def upload(self, url, filename, data=None, formname=None,
-               otherfields=()):
+               otherfields=(), formwrap=True, excepterror=True):
         """Upload a file to the url
 
         :param url:
@@ -270,29 +300,44 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
         """
         if data is None:
             data = open(filename, 'rb')
-        self._upbuffer = io.BytesIO(get_upload_form(
-            filename, data, formname, otherfields))
-        ulheaders = self.stdheaders.copy()
-        ulheaders['Content-Type'] = b'multipart/form-data; boundary=' + BND
-        ulheaders['Content-Length'] = len(uploadforms[filename])
-        self.ulsize = len(uploadforms[filename])
+        ulhdrs = self.stdheaders.copy()
+        if formwrap:
+            self._upbuffer = io.BytesIO(get_upload_form(
+                filename, data, formname, otherfields))
+            ulhdrs['Content-Type'] = b'multipart/form-data; boundary=' + BND
+            ulhdrs['Content-Length'] = len(uploadforms[filename])
+            self.ulsize = len(uploadforms[filename])
+        else:
+            curroff = data.tell()
+            data.seek(0, 2)
+            self.ulsize = data.tell()
+            data.seek(curroff, 0)
+            self._upbuffer = data
+            ulhdrs['Content-Type'] = b'application/octet-stream'
+            ulhdrs['Content-Length'] = self.ulsize
         webclient = self.dupe()
-        webclient.request('POST', url, self._upbuffer, ulheaders)
+        webclient.request('POST', url, self._upbuffer, ulhdrs)
         rsp = webclient.getresponse()
         # peer updates in progress should already have pointers,
         # subsequent transactions will cause memory to needlessly double,
         # but easiest way to keep memory relatively low
-        try:
-            del uploadforms[filename]
-        except KeyError:  # something could have already deleted it
-            pass
+        if formwrap:
+            try:
+                del uploadforms[filename]
+            except KeyError:  # something could have already deleted it
+                pass
         self.rspstatus = rsp.status
-        if rsp.status != 200:
-            raise Exception('Unexpected response in file upload: ' +
-                            rsp.read())
+        if excepterror and (rsp.status < 200 or rsp.status >= 300):
+            raise Exception('Unexpected response in file upload: %s'
+                            % rsp.read())
         body = rsp.read()
         if rsp.getheader('Content-Encoding', None) == 'gzip':
-            body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            try:
+                body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            except IOError:
+                # In case the implementation lied, let the body return
+                # unprocessed
+                pass
         return body
 
     def get_upload_progress(self):
